from pyfyre.widgets.widget import Widget

class Image(Widget):
    """Creates an Image widget.

    Attributes
    ----------
    src : str (positional)
        The link of the image.
    """

    def __init__(self, src, className="", props: dict=None):
        super().__init__("img", className=className, props=props)
        self.src = src
    
    def dom(self):
        element = super().dom()
        element.src = self.src
        
        return element



class Globals:

    # The Parent of all widgets object stored here
    # from `runApp`'s first parameter.
    __PARENT__ = None


    # The current pathname location used for routing.
    __LOC__ = None

    # Dynamic routes
    DYNAMIC_ROUTES = []
    
    # For checking if it's the first time initializing the Router
    PATH_INITIALIZED = False


    # All the event objects for firing.
    __EVENTS__ = {}


    # Path of Assets folder
    __ASSETS__ = ""


from pyfyre.widgets import *

class CounterApp(UsesState):
    def __init__(self):
        self.count = 0

    def increment(self, e):
        self.count = self.count + 1
        self.update()

    def decrement(self, e):
        self.count = self.count - 1
        self.update()

    def build(self):
        return Container(
            className="counter-app",
            children=[
                Text("Counter app example:"),
                Container(
                    className="counters",
                    children=[
                        Button(
                            "-",
                            onClick=self.decrement,
                            className="btn-counter"
                        ),
                        Text(self.count),
                        Button(
                            "+",
                            onClick=self.increment,
                            className="btn-counter"
                        )
                    ]
                )
            ]
        )


from browser import document
from pyfyre.globals import Globals

def runApp(app, mount="app-mount"):
    """This is the main entry point of your app. You must call this
    with an argument object that inherits [UsesState] or just a simple widget.

    Parameters
    ----------
    mount : str
        Where the app will mount on the index.html. This is a provided
        id of a `div` element.
    """

    body = document.getElementById(mount)
    body.innerHTML = ""
    body.appendChild(app.dom())

    Globals.__PARENT__ = app


from pyfyre.widgets import *

class CallToAction(Container):
    def __init__(self, title, desc, link):
        super().__init__(
            className="flex flex-col w-5/6 h-3/6 justify-center mx-auto bg-gradient-to-r from-[#4568dc] to-[#b06ab3] text-white p-10 rounded-3xl my-16",
            children=[
                Text(
                    className="text-md font-bold",
                    textContent=title
                ),
                Text(
                    className="text-3xl w-4/6 font-normal",
                    textContent=desc
                ),
                Container(
                    className="mt-5",
                    children=[
                        Text(
                            className="bg-transparent border border-white w-fit px-5 py-2 text-base rounded-3xl cursor-pointer hover:bg-white hover:text-[#222222]",
                            textContent=f"Visit {link}",
                        )
                    ]
                )
            ]
        )


from browser import document

class Widget:
    
    def __init__ (self, tagname: str, *, className, props: dict=None):
        self.tagname = tagname
        self.className = className
        self.element = None
        self.props = props if props is not None else {}
    
    def dom(self):
        element = document.createElement(self.tagname)
        element.className = self.className

        self.element = element
        
        for key, value in self.props.items():
            element.attrs[key] = value
        
        return element


from pyfyre.widgets.widget import Widget

class BaseException:
    def __init__(self, e):
        self.e = e

    def dom(self):
        return self.TextException(self.e).dom()

    class TextException(Widget):
        def __init__(self, error: str, className="", props: dict=None):
            super().__init__("h1", className=className, props=props)
            self.error = error
        
        def dom(self):
            element = super().dom()
            element.textContent = self.error
            
            element.attrs["style"] = "background-color: #efa3a3; width: fit-content; padding: 10px;"
            
            return element

class RenderError(BaseException):
    pass


from pyfyre.widgets import *
from src.components.cta import CallToAction
from src.components.description import Descriptions

# You can make components by inherting a Widget class
# and initialize it on the `super().__init__()`.
class HomePage(Container):
    def __init__(self):
        
        _features = [
            [
                "Developer friendly",
                "https://upload.wikimedia.org/wikipedia/commons/thumb/c/c3/Python-logo-notext.svg/1200px-Python-logo-notext.svg.png",
                "The language you already know.",
                "PyFyre allows you to develop apps with the language you already know, Python. Create reactive websites fast with ease. Build your apps in minutes, not days.",
                "Show me"
            ],
            [
                "Fast",
                "https://i.imgur.com/yI8tUab.png",
                "Fast transpilation time on the fly.",
                "PyFyre transpiles your code into small, vanilla JS that works for every browser with the power of Brython that stays fast, and always will be.",
                "Show me"
            ],
            [
                "Productive",
                "https://pbs.twimg.com/profile_images/909757546063323137/-RIWgodF_400x400.jpg",
                "Use Python Packages on the web.",
                "Use CPython's existing packages on PyPi.org/, develop sites quickly and efficiently with a large selection of packages.",
                "Visit pypi.org/"
            ],
        ]

        layouts = []

        for i in range(len(_features)):
            layout_1 = "md:flex-row-reverse" if i % 2 == 0 else "md:flex-row"
            layout_2 = "items-start" if i % 2 == 0 else "items-end"

            _layouts = []
            _layouts.append(layout_1)
            _layouts.append(layout_2)
            
            layouts.append(_layouts)

        def feature(i):
            return Container(
                className="",
                children=[
                    Container(
                        className=f"flex flex-col items-center px-5 py-16 mx-auto {layouts[i][0]} lg:px-28",
                        children=[
                            Container(
                                className="w-full lg:w-1/3 lg:w-3/12 md:w-1/2",
                                children=[
                                    Image(
                                        className="object-cover object-center rounded-lg",
                                        src=_features[i][1]
                                    )
                                ]
                            ),
                            Container(
                                className=f"flex flex-col {layouts[i][1]} mb-16 text-left lg:flex-grow md:w-1/2 lg:pr-24 md:pr-16 md:mb-0",
                                children=[
                                    Text(
                                        className="mb-8 text-lg font-semibold tracking-widest text-black uppercase title-font",
                                        textContent=_features[i][0]
                                    ),
                                    Text(
                                        className="mb-8 text-2xl font-black tracking-tighter text-black md:text-5xl title-font",
                                        textContent=_features[i][2]
                                    ),
                                    Text(
                                        className="mb-8 text-base leading-relaxed text-left text-blueGray-600",
                                        textContent=_features[i][3]
                                    ),
                                    Container(
                                        className="flex flex-col justify-center lg:flex-row",
                                        children=[
                                            Text(
                                                className="flex items-center px-6 py-2 mt-auto font-semibold text-white transition duration-500 ease-in-out transform bg-blue-600 rounded-lg hover:bg-blue-700 focus:shadow-outline focus:outline-none focus:ring-2 ring-offset-current ring-offset-2",
                                                textContent=_features[i][4]
                                            ),
                                        ]
                                    )
                                ]
                            )
                        ]
                    ),
                ]
            )

        super().__init__(
            className = "flex flex-col h-screen w-full",
            children = [
                Container(
                    className="flex flex-col w-full h-5/6 bg-[#222222]",
                    children=[
                        Container(
                            className="flex flex-col w-full mt-28",
                            children=[
                                Text(
                                    className = "text-3xl mx-auto font-bold bg-clip-text h-20 text-transparent bg-gradient-to-r from-cyan-500 to-blue-500 xl:text-6xl",
                                    textContent = "Build reactive static UI using Python."
                                ),
                                Text(
                                    className = "px-28 text-xl mx-auto mt-5 xl:px-52 text-center text-white",
                                    textContent = "PyFyre is a fast, declarative, and incrementally adoptable Python web frontend framework for building reactive static single-page web applications easily and efficiently."
                                )
                            ]
                        ),
                        Container(
                            className="flex space-x-3 mx-auto mt-20 mb-44",
                            children=[
                                Link(
                                    className="bg-[#444444] w-fit px-5 py-2 text-base rounded-3xl text-white cursor-pointer",
                                    textContent="Documentation",
                                    to="/docs"
                                ),
                                Link(
                                    className="bg-[#fab327] w-fit px-5 py-2 text-base rounded-3xl text-white cursor-pointer",
                                    textContent="Get started",
                                    to="/docs/introduction"
                                )
                            ]
                        )
                    ]
                ),
                ListBuilder(
                    className="",
                    count=len(_features),
                    builder=feature
                ),
                CallToAction("Brython", "PyFyre is powered by Brython, a Python 3 implementation for client-side web programming.", "brython.info"),
            ]
        )


from pyfyre.widgets import *
from pyfyre.globals import Globals
from src.components.docs_pages.getstarted import *
from src.components.docs_pages import DocsPage

# You can make components by inherting a Widget class
# and initialize it on the `super().__init__()`.
class Docs(Container):
    def __init__(self):

        pages = {
            "/": Introduction,
            "/?": NotFinished,
            "/docs": Introduction,
            "/docs/introduction": Introduction,
            "/docs/introduction": Introduction,
            "/docs/installation": Installation
        }

        currentPage = pages[Globals.__LOC__] if Globals.__LOC__ is not None else pages["/docs/introduction"]

        super().__init__(
            className = "flex flex-row h-full w-full overflow-y-auto",
            children = [
                Sidebar(),
                currentPage()
            ]
        )

class NotFinished(DocsPage):
    def __init__(self):
        
        def content():
            return Text(
                className="",
                textContent="This documentation page is under construction yet. Please come back later."
            )

        super().__init__(content)

class Sidebar(Container):
    def __init__(self):

        docs = {
            "Get started": {
                "Introduction": [
                    ["What and why PyFyre?", "/docs/introduction"],
                    ["Python on the client-side web", "/docs/introduction"],
                ],
                "Installation": [
                    ["Install", "/docs/installation"],
                ],
            },
            "Development": [
                ["Introduction to components", "/?"],
                ["Building components", "/?"],
                ["Component properties", "/?"],
                ["Navigation & Routing", "/?"],
                ["Python Packages on the web", "/?"],
            ],
            "Cookbook": [
                ["Increment-decrement", "/?"],
                ["Display images", "/?"],
                ["Create a list from Python List", "/?"],
                ["Navigation", "/?"]
            ]
        }

        docs_list = list(docs)

        def document(i):

            part_name = docs_list[i]
            contents = docs[docs_list[i]]

            def contentBuilder(contentIndex):
                content = None

                if type(contents) == dict:
                    content_list = list(contents)
                    content_contents = contents[content_list[contentIndex]]

                    title = content_list[contentIndex]
                    
                    content = content_contents

                    def nestedDocument(i):
                        return Link(
                            className="text-base ml-6",
                            textContent=f"{i+1}. {content[i][0]}",
                            to=content[i][1]
                        )

                    return Container(
                        className="mt-4",
                        children=[
                            Text(
                                className="text-base",
                                textContent=title
                            ),
                            ListBuilder(
                                className="flex flex-col",
                                count=len(content),
                                builder=nestedDocument
                            )
                        ]
                    )

                else:
                    content = contents[contentIndex][0]

                    return Link(
                        className="text-base",
                        textContent=content,
                        to=contents[contentIndex][1]
                    )
        
            return Container(
                className="",
                children=[
                    Text(
                        className="font-bold text-lg",
                        textContent=f"{part_name}"
                    ),
                    ListBuilder(
                        className="flex flex-col",
                        count=len(contents),
                        builder=contentBuilder
                    )
                ]
            )

        super().__init__(
            className="bg-[#f2f2f2] h-screen overflow-y-auto w-80 p-10 text-[#222222] fixed border-r",
            children=[
                ListBuilder(
                    className="flex flex-col space-y-6 mb-10",
                    count=len(docs),
                    builder=document
                )
            ]
        )


from pyfyre.globals import Globals
from pyfyre.globals.events import Events
from pyfyre.pyfyre import runApp

from browser import document, window, bind

class Router:
    """Router v0.3-alpha.
    
    Creates a wrapper object with window location
    history listener that when it's changed,
    the whole app rerenders and get the provided
    routes and map it correctly.

    Attributes
    ----------
    routes : dict
        A dictionary of object routes. Every value
        must inherit the [UsesState] object.

        For instance:
            Router(
                routes={
                    `"/": Home()`,
                    `"/about": About()`,

                    # Dynamic routing, `<>` tells the Router that the `about` path has dynamic routes.
                    `"/about/<>: AboutPerson()`
                }
            )

    query : str
        If dynamic routing is allowed to the route, you can
        get the query of the route.

        For instance:
        `https://www.nicebookies.com/users/users_cool_username`

        -----

        from pyfyre.router import Router

        Router(
            routes={
                `"/users/<>": UserProfile()`,
            }
        )

        -----
        
        from pyfyre.router import Router

        class UserProfile(UsesState):
            def __init__(self):
                self.username = Router.query

            def build(self):
                return Text(f"Username: {self.username}")

    """

    def __init__(self, routes):
        self.routes: dict = routes

        if not Globals.PATH_INITIALIZED:
            Globals.__LOC__ = window.location.pathname
            Globals.PATH_INITIALIZED = True

        if not "change_route" in Globals.__EVENTS__:
            Events.add("change_route")

            self.listenRoute()

        for routeName, view in self.routes.items():
            pathname = routeName.split('/')

            query = pathname[-1]
            pathurl = pathname[-2]

            if query == "<>":
                Globals.DYNAMIC_ROUTES.append([pathurl, view])
        
    def dom(self):

        try:
            dom = self.routes[Globals.__LOC__].dom()
        except KeyError:
            try:
                pathurl = self.pathurl

                for route in Globals.DYNAMIC_ROUTES:
                    if route[0] == pathurl:
                        dom = route[1].dom()
            except Exception:
                # TODO: Create a 404.py that allows developers to customize 404 UI.
                # that can also be modified on the settings so they can
                # customize the 404.py name for their like.
                raise Exception("PATH 404: Cannot find the path.")

        return dom

    def listenRoute(self):
        def changeRoute():
            runApp(Globals.__PARENT__)

        Events.addListener("change_route", changeRoute)

    @property
    def query(self):
        path = f"{Globals.__LOC__}"
        pathname = path.split('/')
        query = pathname[-1]

        return query

    @property
    def pathurl(self):
        path = f"{Globals.__LOC__}"
        pathname = path.split('/')
        pathurl = pathname[-2]

        return pathurl

    @staticmethod
    def push(location):
        Globals.__LOC__ = location

        Events.broadcast("change_route")
        window.history.pushState(None, None, location)

        @bind(window, 'popstate')
        def popState(e):
            Globals.__LOC__ = window.location.pathname
            Events.broadcast("change_route")
            e.preventDefault()


from pyfyre.widgets.widget import Widget

class ListBuilder(Widget):
    """Creates a ListBuilder widget.

    This is like for loops in programming languages,
    you provide a range (count for instance)
    and the content of the loop (builder for instance).

    The widget returned on the [builder] will be appended
    to the DOM for [count] times.

    Attributes
    ----------
    count : int
        How many times will the ListBuilder [build] method
        will run?
    builder : method -> Widget
        The widget that will append in [count] times.
    """
    
    def __init__(self, count=1, builder=None, className="", props: dict=None):
        super().__init__("div", className=className, props=props)
        self.count = count
        self.builder = builder
    
    def dom(self):
        element = super().dom()
        
        for i in range(self.count):
            element.appendChild(self.builder(i).dom())
        
        return element



from pyfyre.widgets.button import Button
from pyfyre.widgets.container import Container
from pyfyre.widgets.image import Image
from pyfyre.widgets.link import Link
from pyfyre.widgets.listbuilder import ListBuilder
from pyfyre.widgets.states import UsesState
from pyfyre.widgets.text import Text

__all__ = [
    'Button', 'Container', 'Image',
    'Link', 'ListBuilder', 'UsesState',
    'Text'
]


from pyfyre.widgets import *
from src.components.docs_pages import DocsPage
from src.components.docs_pages.section import Section

class Introduction(DocsPage):
    def __init__(self):

        def whatIsPyFyre():
            return Container(
                className="",
                children=[
                    Text(
                        className="mt-5",
                        textContent = "PyFyre is a fast, declarative, and incrementally adoptable Python web frontend framework for building reactive static web applications easily and efficiently. It allows you to create frontend applications with the programming language you already know, Python. Thanks to Brython's powerful tooling, PyFyre also allows you to use existing Python Packages on the client-side web. PyFyre also fits to create Single-Page applications with powerful built-in tools like PyFyre Router, and more."
                    ),
                    Text(
                        className="mt-4",
                        textContent = "PyFyre has a quite of mix with the existing largest framework, React and Flutter. PyFyre is inspired by Flutter's beautiful development experience and React's speed and efficiency. Together, PyFyre is a developer and user friendly framework in Python."
                    ),
                ]
            )

        def pythonOnClientSideWeb():
            return Container(
                className="",
                children=[
                    Text(
                        className="mt-5",
                        textContent = "You're probably new with Python on the client-side web as the current frontend frameworks used today are made with Node.js Javascript. PyFyre's mission is not to *really* reinvent the wheel, its mission is to allow the frontend web to access existing Python Packages."
                    ),
                    Text(
                        className="mt-4",
                        textContent = "How? PyFyre compiles your Python code in Javascript itself instead of CPython with the help of Brython (Browser Python). It's still the same, just not Javascript but it's Python. Thanks to Brython, you can also access vanilla Javascript."
                    ),
                    Text(
                        className="mt-4",
                        textContent = "PyFyre is in the top of Brython, most of its operations are on Brython. PyFyre just helps developers to create apps easily just like React, Vue, and Flutter."
                    ),
                ]
            )
        
        def content():
            return Container(
                className="text-base",
                children=[
                    Section(
                        "What and why PyFyre",
                        content=whatIsPyFyre
                    ),
                    Container(className="mt-20", children=[]),
                    Section(
                        "Python on the client-side web",
                        content=pythonOnClientSideWeb
                    ),
                ]
            )

        super().__init__(content=content)


class Installation(DocsPage):
    def __init__(self):

        def install():
            return Container(
                className="",
                children=[
                    Text(
                        className="mt-5",
                        textContent = "PyFyre installation is really straightforward, you only need to be familiar of Python and CSS, and a working environment."
                    ),
                    Text(
                        className="mt-4",
                        textContent = "To install PyFyre, you need to install Python 3.x, and a working `pip` command. To make sure you installed it correctly, go to your terminal (cmd on Windows) and type `py --version` if Python version shows, you already have Python. To check `pip`, type `pip --version` and it should show as well."
                    ),

                    # Install PyFyre CLI
                    Text(
                        className="mt-7 text-xl font-medium",
                        textContent = "Install PyFyre CLI"
                    ),
                    Text(
                        className="mt-4",
                        textContent = "You need PyFyre CLI to create projects, generate application and perform commands."
                    ),
                    Text(
                        className="mt-4",
                        textContent = "To install the PyFyre CLI, open a terminal window (cmd on Windows) and run the following command:"
                    ),
                    Text(
                        className="mt-2 bg-[#f2f2f2] px-4 py-4 rounded-md",
                        textContent = "pip install pyfyre"
                    ),

                    # Create an App
                    Text(
                        className="mt-7 text-xl font-medium",
                        textContent = "Create an App"
                    ),
                    Text(
                        className="mt-4",
                        textContent = "Start your app and create a new workspace to develop your apps."
                    ),
                    Text(
                        className="mt-4",
                        textContent = "To create a new app, run the PyFyre CLI command:"
                    ),
                    Text(
                        className="mt-2 bg-[#f2f2f2] px-4 py-4 rounded-md",
                        textContent = "pyfyre create-app my-cool-app"
                    ),
                    Text(
                        className="mt-4",
                        textContent = "You can replace `my-cool-app` with the name of the app you desired."
                    ),

                    # Run the App
                    Text(
                        className="mt-7 text-xl font-medium",
                        textContent = "Run the App"
                    ),
                    Text(
                        className="mt-4",
                        textContent = "PyFyre has a development server that has hot reload that detects changes in your code and reload the application automatically."
                    ),
                    Text(
                        className="mt-4 ml-2",
                        textContent = "1. Navigate to the app folder, in our case, it's `my-cool-app`"
                    ),
                    Text(
                        className="mt-4 ml-2",
                        textContent = "2. Run PyFyre CLI command:"
                    ),
                    Text(
                        className="mt-2 bg-[#f2f2f2] px-4 py-4 rounded-md",
                        textContent = "pyfyre runapp"
                    ),
                    Text(
                        className="mt-4",
                        textContent = "And voila! You now have a running PyFyre app on your local machine."
                    ),
                    Text(
                        className="mt-4",
                        textContent = "Note: Once you have run the app, PyFyre will generate a folder called `__serve__`, the folder contains all the files the running server need. You can delete it once `runapp` has finished or you can ignore it on .gitignore file."
                    ),
                ]
            )
        
        def content():
            return Container(
                className="text-base",
                children=[
                    Section(
                        "Install",
                        content=install
                    ),
                ]
            )

        super().__init__(content=content)


from pyfyre.widgets import *

class Section(Container):
    def __init__(self, title, content):
        super().__init__(
            className="text-base",
            children=[
                Text(
                    className="text-4xl font-bold",
                    textContent=title
                ),
                content()
            ]
        )


from pyfyre.widgets import *

class DocsPage(Container):
    def __init__(self, content):
        super().__init__(
            className="text-[#222222] px-52 py-10 ml-52",
            children=[
                content()
            ]
        )


from pyfyre.widgets.widget import Widget

class Container(Widget):
    """Creates a Container Widget.

    This is like `div` in HTML.

    Attributes
    ----------
    children : list<Widget>
        The list of children of the Container widget.
    """
    
    def __init__(self, children=[], className="", props: dict=None):
        super().__init__("div", className=className, props=props)
        self.children = children
    
    def dom(self):
        element = super().dom()
        
        for child in self.children:
            element.appendChild(child.dom())
        
        return element


from pyfyre.widgets import *

class Descriptions(Container):
    def __init__(self):
        super().__init__(
            className="",
            children=[
                Image(),
                Container(
                    children=[
                        Text(),
                        Text(),
                        Text()
                    ]
                )
            ]
        )


from pyfyre.widgets.widget import Widget

class Text(Widget):
    """Creates a Text widget.

    This is a simple widget to put a text on the screen.

    Attributes
    ----------
    textContent : str
        The text content of the Text widget
    """
    
    def __init__(self, textContent: str, className="", props: dict=None):
        super().__init__("p", className=className, props=props)
        self.textContent = textContent
    
    def dom(self):
        element = super().dom()
        element.textContent = self.textContent
        
        return element


from pyfyre.widgets import *
from pyfyre.pyfyre import runApp
from src.main import HomePage
from src.components.header import Header
from src.components.docs import Docs

class App(UsesState):
    def __init__(self):
        self.greet = "Hello, there"

    def build(self):
        return Router(
            routes={
                "/": Home(),
                "/?": Documentation(),
                "/docs": Documentation(),
                "/docs/introduction": Documentation(),
                "/docs/installation": Documentation(),
            }
        )
        
class Home(Container):
    def __init__(self):
        super().__init__(
            className = "home",
            children = [
                Header(),
                HomePage()
            ]
        )

class Documentation(Container):
    def __init__(self):
        super().__init__(
            className="docs",
            children=[
                Header(),
                Docs()
            ]
        )

runApp(
    App(),
    mount="app-mount"
)


"""PyFyre States"""

from pyfyre.widgets.widget import Widget
from pyfyre.core.exceptions import RenderError

class UsesState:
    """Create a component that is stateful.

    If you think the component will update again and again,
    you might want to use [UsesState] for that component.

    This allows you to rerender the component when data
    changes by calling `self.update()` method.

    This also wraps the uncaught errors on `self.onerror` method
    where you can also override and create your own UI error.

    Your [UsesState] method must override the build method returning
    an another component or a Widget.
    """

    def __init__(self):
        self.domElement = None
    
    def build(self):
        pass

    def dom(self):
        try:
            self.domElement = self.build().dom()
            return self.domElement
        except Exception as e:
            self.domElement = self.onerror(e).dom()
            return self.domElement

    def onerror(self, e):
        """
        Handle errors on every component separately. When an error
        interrupts the `build` method on the component, this `onerror`
        method will be fired and you can handle it professionally without
        blowing the entire app.
        
        Override this method by declaring another `onerror` method on your
        component and taking another parameter where the error exception will
        pass. You must return a widget in which the UI you want your
        users to see temporarily.
        """
        
        print("ERR!", e)

        return RenderError("Oh no! Something went wrong. We're working on fixing it.")

    def update(self):
        """
        Updates the DOM element. Instead of painting the app again
        for state change, only update the DOM where the change happens.
        """
        
        parentNode = self.domElement.parentNode
        self.domElement.remove()
        self.domElement = self.dom()
        parentNode.appendChild(self.domElement)